#include <cpar.h>
#include <cpar_internal.h>

const char *cpar_type_names[CPAR_BASIC_TYPES] =
  {
    "CPAR_UNKNOWN",
    "CPAR_BOOL",
    "CPAR_INT",
    "CPAR_UINT",
    "CPAR_DOUBLE",
    "CPAR_REAL",
    "CPAR_STRING",
    "CPAR_BOOL_POINTER",
    "CPAR_CHAR_POINTER",
    "CPAR_INT_POINTER",
    "CPAR_UINT_POINTER",
    "CPAR_DOUBLE_POINTER",
    "CPAR_REAL_POINTER",
    "CPAR_POINTER",
    "CPAR_GPOINTER" 
  };

// _cpar_last_pointer_type is the last pointer type assigned using _cpar_assign_pointer_type()
uint _cpar_last_pointer_type = CPAR_CUSTOM_GPOINTER_TYPE-1;

// cpar_empty_table is an empty cpar object so that you can pass an empty cpar table
// without having to create and free one.
// (Disabled: this is too dangerous without the ability to write-protect a table.
//  The side-effects could be nasty.)
//cpar_table *cpar_empty_table = NULL;

// ==============================================================
// Initialize the global variables in the library when the library
// is first loaded.  This uses a non-standard gcc extension
// (the __attribute__((constructor)) part).  If this is ported to
// new compilers, you should use your compiler's method for
// calling a library initializer.
// ==============================================================
static void __attribute__((constructor)) _cpar_init(void)
{
// initialize global variable _cpar_abort_function to internal version
_cpar_abort_function = _cpar_internal_abort_function;
//cpar_empty_table = CPAR_CREATE_EMPTY_PARAMETERS();
}

// ==============================================================
// Clean up when library unloads
// ==============================================================
static void __attribute__((destructor)) _cpar_end(void)
{
//CPAR_FREE(cpar_empty_table);
}

// ==============================================================
// Return 1 if the given type is a gpointer custom type
// ==============================================================
bool _cpar_type_is_gpointer_custom(uint type)
{
  if (type >= CPAR_CUSTOM_GPOINTER_TYPE && type <= _cpar_last_pointer_type)
    return 1;
  else
    return 0;
}

// ==============================================================
// Return 1 if the given type is with type limits and not a custom gpointer
// ==============================================================
bool _cpar_type_is_normal(uint type)
{
  if ( type >= CPAR_INT && type <= CPAR_GPOINTER )
    return 1;
  else
    return 0;
}

// ==============================================================
// Abort if the given type is not a possible type
// ==============================================================
void _cpar_assert_type(char *file, uint line, char *function, uint type)
{
  if ( ! _cpar_type_is_normal(type) && ! _cpar_type_is_gpointer_custom(type) )
      _cpar_abort(file, line, function, "type %d is not a valid cpar type", type);
}

// ==============================================================
// Abort if the pointer doesn't point to a cpar_table
// ==============================================================
void _cpar_assert_cpar_table(char *file, uint line, cpar_table *vp)
{
if (vp == NULL) { _cpar_abort(file, line, "Pointer to NULL is not a cpar_table"); }
if ((*vp)->magic_key == _CPAR_MAGIC_KEY_DELETED)
  {
    _cpar_abort(file, line, "Pointer to %lX is a cpar_table that was freed", (unsigned long int)*vp);
  }
if ((*vp)->magic_key == _CPAR_MAGIC_KEY) { return; }
else
  {
    _cpar_abort(file, line, "Pointer to %lX is not a cpar_table", (unsigned long int)*vp);
  }
printf("should not reach here\n");
}

// ==============================================================
// GHFunc for the function immediately below
// ==============================================================
void _cpar_merge_cpar_tables_GHFunc(gpointer from_key_pointer, gpointer value, gpointer user_data)
{
char *from_key = (char *)from_key_pointer;
cpar_parameter_struct *from_parameter = (cpar_parameter_struct *)value;
ghfunc_passer_struct *ghfunc_passer = (ghfunc_passer_struct *)user_data;
cpar_table *to_vp = ghfunc_passer->to_vp;
char *to_key = ghfunc_passer->to_key;
if (to_key == NULL) { to_key = from_key; }
uint merge_mode = ghfunc_passer->mode;

// Check to_vp
_cpar_assert_cpar_table(ghfunc_passer->file, ghfunc_passer->line, to_vp);

// Get the pointer if this to_key is already in the list
#ifdef CPAR_DEBUG
    printf("_cpar_merge_cpar_tables_GHFunc: merge_mode=%d, from_key = %s, to_key = %s\n", 
           merge_mode, from_key, to_key);
#endif
gpointer lookup_pointer = g_hash_table_lookup((*to_vp)->hash_table, to_key);

// Check for the special case of a default parameter being merged into a
// non-default parameter, which should always result in the merge failing,
// regardless of the merge mode.
if (lookup_pointer != NULL)
  {
    cpar_parameter_struct *to_parameter = (cpar_parameter_struct *)lookup_pointer;
    if (to_parameter->is_default == false && from_parameter->is_default == true) return;
  }

// Check for the special case of a non-default parameter being merged into a
// default parameter, which should alway succeed, regardless of the merge mode
if (lookup_pointer != NULL)
  {
    cpar_parameter_struct *to_parameter = (cpar_parameter_struct *)lookup_pointer;
    if (to_parameter->is_default == true && from_parameter->is_default == false)
      {
        goto _cpar_merge_cpar_tables_GHFunc_MERGE;
      }
  }

switch(merge_mode)
  {
    case CPAR_MERGE_NO_OVERWRITE_MODE:
      // fail if key exists and it is not a default parameter
      if (lookup_pointer != NULL && from_parameter->is_default == false)
        {
          _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, 
                      "key \"%s\" already exists in hash during merge", to_key);
        }
      break;

    case CPAR_MERGE_OVERWRITE_MODE:
      // overwrite if key exists (just always goes)
      // Delete first if it's already there.  This shouldn't be necessary according to the
      // glib docs, but I think maybe there's a bug... I was seeing the same key going into
      // the hash table multiple times using the g_hash_table_insert function.
      if (lookup_pointer != NULL)
        {
          //g_hash_table_remove(to_vp->hash_table, to_key);
        }
      break;

    case CPAR_MERGE_NEW_MODE:
      // only add new keys (but just skip instead of fail if not new)
      if (lookup_pointer != NULL) return;
      break;

    case CPAR_MERGE_UPDATE_MODE:
      // only overwrite if key exists
      if (lookup_pointer == NULL) return;
      //g_hash_table_remove(to_vp->hash_table, to_key); // see above for why this is needed
      break;

    default:
      _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, "unknown mode: %d", merge_mode);
      break;
  } // end switch()

_cpar_merge_cpar_tables_GHFunc_MERGE:;
cpar_parameter_struct *to_parameter =
    _cpar_create_cpar_parameter_struct(ghfunc_passer->file, ghfunc_passer->line, to_key, 
                                       &(from_parameter->value.gpointer_value), from_parameter);
// Allocate a new key
char *new_key = malloc(strlen(to_key)+1);
strcpy(new_key, to_key);
g_hash_table_insert((*to_vp)->hash_table, new_key, (gpointer)to_parameter);
} // end _cpar_merge_cpar_tables_GHFunc()

// ==============================================================
// Add the contents of the second parameter table to the first
// ==============================================================
void _cpar_merge_cpar_tables(char *file, uint line, cpar_table *to_vp, cpar_table *from_vp, uint mode)
{
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.mode = mode;
ghfunc_passer.to_vp = to_vp;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.to_key = NULL; // Just copy keys from old to new
_cpar_assert_cpar_table(file, line, to_vp);
_cpar_assert_cpar_table(file, line, from_vp);
g_hash_table_foreach((*from_vp)->hash_table, _cpar_merge_cpar_tables_GHFunc, (gpointer)(&ghfunc_passer));
} // end _cpar_merge_cpar_tables()


// ==============================================================
// Keep only listed parameters in parameter table
// ==============================================================
void _cpar_keep_parameters_by_key(char *file, uint line, cpar_table *vp, ...)
{
// Copy values of "keep" key list to a new hash table
va_list args;
char *key;
gpointer value;
cpar_parameter_struct *old_parameter;
cpar_parameter_struct *new_parameter;
cpar_table *new_vp;

if (vp == NULL) return;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

// Create a new parameter table
new_vp = _cpar_create_empty_cpar_table(file, line);
  
va_start(args, vp);
key = va_arg(args, char *);
while (key != NULL)
  {
    value = g_hash_table_lookup((*vp)->hash_table, key);
    if (value != NULL)
      {
        old_parameter = (cpar_parameter_struct *)value;
        // New parameter must be allocated because all the old ones are going to be
        // deleted when the whole table is freed.
        // Allocate a new key
        char *new_key = malloc(strlen(key)+1);
        strcpy(new_key, key);
        new_parameter = _cpar_create_cpar_parameter_struct(file, line, key, 
                                                           &(old_parameter->value.gpointer_value), 
                                                           old_parameter);
        g_hash_table_insert((*new_vp)->hash_table, new_key, (gpointer)new_parameter);
      }
    key = va_arg(args, char *);
  } // end while

// Get rid of the old parameter table
_cpar_destroy_cpar_table(file, line, vp);
// Point to new parameter table
*vp = *new_vp;
} // end _cpar_keep_parameters_by_key()


// ==============================================================
// Remove listed parameters from table
// keys do not have to exist in the table
// ==============================================================
void _cpar_remove_parameters_by_key(char *file, uint line, cpar_table *vp, ...)
{
va_list args;
char *key;
gpointer value;
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.vp = vp;

if (vp == NULL) return;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

va_start(args, vp);
key = va_arg(args, char *);
while (key != NULL)
  {
    value = g_hash_table_lookup((*vp)->hash_table, key);
    if (value != NULL)
      {
        g_hash_table_remove((*vp)->hash_table, key);
      }
    key = va_arg(args, char *);
  }
} // end _cpar_remove_parameters_by_key()


// ==============================================================
// Return 1 if the beginning of this parameter matches the string ghfunc_passer->compare
// ==============================================================
gboolean _cpar_select_parameters_by_string_GHRFunc(gpointer key, gpointer value, gpointer user_data)
{
ghfunc_passer_struct *ghfunc_passer = (ghfunc_passer_struct *)user_data;

guint compare_length = strlen(ghfunc_passer->compare_string);
return (! strncmp(ghfunc_passer->compare_string, (char *)key, compare_length));
} // end _cpar_select_parameters_by_string_GHRFunc()


// ==============================================================
// Return the logical opposite of _cpar_select_parameters_by_string_GHRFunc()
// ==============================================================
gboolean _cpar_deselect_parameters_by_string_GHRFunc(gpointer key, gpointer value, gpointer user_data)
{
return (! _cpar_select_parameters_by_string_GHRFunc(key, value, user_data));
} // end _cpar_deselect_parameters_by_string_GHRFunc()


// ==============================================================
// Remove or keep parameters whose keys don't start with string
// ==============================================================
void _cpar_change_parameters_by_string(char *file, uint line, cpar_table *vp, char *string, GHRFunc func)
{
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.vp = vp;
ghfunc_passer.compare_string = string;

if (vp == NULL) return;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

g_hash_table_foreach_remove((*vp)->hash_table, func, (gpointer)(&ghfunc_passer));
} // end _cpar_change_parameters_by_string()


// ==============================================================
// Remove parameters whose keys don't start with string
// ==============================================================
void _cpar_remove_parameters_by_string(char *file, uint line, cpar_table *vp, char *string)
{
_cpar_change_parameters_by_string(file, line, vp, string, _cpar_select_parameters_by_string_GHRFunc);
} // end _cpar_remove_parameters_by_string()


// ==============================================================
// Keep parameters whose keys start with string
// ==============================================================
void _cpar_keep_parameters_by_string(char *file, uint line, cpar_table *vp, char *string)
{
_cpar_change_parameters_by_string(file, line, vp, string, _cpar_deselect_parameters_by_string_GHRFunc);
} // end _cpar_remove_parameters_by_string()


// ==============================================================
// The GHFunc for the list function below
// ==============================================================
void _cpar_list_parameters_GHFunc(gpointer key, gpointer value, gpointer user_data)
{
cpar_parameter_struct *parameter = (cpar_parameter_struct *)value;
ghfunc_passer_struct *ghfunc_passer = (ghfunc_passer_struct *)user_data;
uint print_type;

#ifdef CPAR_DEBUG
  fprintf(CPAR_ERROR_STREAM, "key: \"%s\", location = %lX, ", (char *)key, (unsigned long int)(parameter));
#else
  fprintf(CPAR_ERROR_STREAM, "key: \"%s\", ", (char *)key);
#endif

if (parameter->is_default == true) fprintf(CPAR_ERROR_STREAM, "default, ");

print_type = parameter->type;
if (_cpar_type_is_gpointer_custom(print_type)) print_type = CPAR_GPOINTER;
switch(print_type)
  {
    case CPAR_BOOL:
      fprintf(CPAR_ERROR_STREAM, "type: BOOL (type %d), ", parameter->type);
      if (parameter->value.bool_value) fprintf(CPAR_ERROR_STREAM, "value: TRUE\n");
      else fprintf(CPAR_ERROR_STREAM, "value: FALSE\n");
    break;
    case CPAR_INT:
      fprintf(CPAR_ERROR_STREAM, "type: INT (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %d\n", parameter->value.int_value);
    break;
    case CPAR_UINT:
      fprintf(CPAR_ERROR_STREAM, "type: UINT (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %d\n", parameter->value.uint_value);
    break;
    case CPAR_DOUBLE:
      fprintf(CPAR_ERROR_STREAM, "type: DOUBLE (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %g\n", parameter->value.double_value);
    break;
    case CPAR_REAL:
      fprintf(CPAR_ERROR_STREAM, "type: REAL (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %g (converted to double)\n", (double)(parameter->value.real_value));
    break;
    case CPAR_BOOL_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: BOOL POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.bool_pointer_value));
    break;
    case CPAR_STRING:
      fprintf(CPAR_ERROR_STREAM, "type: STRING (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: \"%s\"\n", parameter->value.char_pointer_value);
    break;
    case CPAR_CHAR_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: CHAR POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.char_pointer_value));
    break;
    case CPAR_INT_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: INT POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.int_pointer_value));
    break;
    case CPAR_UINT_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: UINT POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.uint_pointer_value));
    break;
    case CPAR_DOUBLE_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: DOUBLE POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.double_pointer_value));
    break;
    case CPAR_REAL_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: REAL POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.real_pointer_value));
    break;
    case CPAR_CPAR_POINTER:
      fprintf(CPAR_ERROR_STREAM, "type: CPAR POINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.cpar_pointer_value));
    break;
    case CPAR_GPOINTER:
      fprintf(CPAR_ERROR_STREAM, "type: GPOINTER (type %d), ", parameter->type);
      fprintf(CPAR_ERROR_STREAM, "value: %lX\n", (unsigned long int)(parameter->value.gpointer_value));
    break;
    default:
      _cpar_abort(ghfunc_passer->file, ghfunc_passer->line, "unexpected parameter type == %d",
          parameter->type);
    break;
  } // end switch
} // end _cpar_list_parameters_GHFunc()

// ==============================================================
// List all the parameters in a cpar_table
// ==============================================================
void _cpar_list_parameters(char *file, uint line, cpar_table *vp)
{
// Pass the file and line number in this struct
ghfunc_passer_struct ghfunc_passer;

if (vp == NULL) return;
ghfunc_passer.file = file;
ghfunc_passer.line = line;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

fprintf(CPAR_ERROR_STREAM, "------------------------\n");
g_hash_table_foreach((*vp)->hash_table, _cpar_list_parameters_GHFunc, (gpointer)(&ghfunc_passer));
fprintf(CPAR_ERROR_STREAM, "------------------------\n");
fflush(CPAR_ERROR_STREAM);
} // end _cpar_list_parameters()


// ==============================================================
// Return 1 if all parameters exist, 0 otherwise
// ==============================================================
bool _cpar_exists(char *file, uint line, cpar_table *vp, ...)
{
va_list variable_parameters;
char *key;
gpointer value;
uint return_value = 0;

if (vp == NULL) { return 0; }

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

va_start(variable_parameters, vp);
key = va_arg(variable_parameters, char *);
while (key != NULL)
  {
    value = g_hash_table_lookup((*vp)->hash_table, key);
    if (value == NULL)
      {
        goto _cpar_exists_END;
      }
    key = va_arg(variable_parameters, char *);
  }
return_value = 1;

_cpar_exists_END:
va_end(variable_parameters);
return return_value;
} // end _cpar_exists()


// ==============================================================
// The GHFunc for the list function below
// Return true (and thus delete the parameter) from the table if the default flag is set
// ==============================================================
gboolean _cpar_delete_defaults_GHFunc(gpointer key, gpointer value, gpointer user_data)
{
gboolean return_value;
ghfunc_passer_struct *ghfunc_passer = (ghfunc_passer_struct *)user_data;
cpar_table *vp = ghfunc_passer->vp;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(ghfunc_passer->file, ghfunc_passer->line, vp);

if (CPAR_IS_DEFAULT(vp, false/*verbose*/, key))
  {
    #ifdef CPAR_DEBUG
    printf("_cpar_delete_defaults_GHFunc: key \"%s\" will be removed\n", (char *)key);
    #endif
    return_value = TRUE;
  }
else return_value = FALSE;
return return_value;
} // end _cpar_delete_defaults_GHFunc()


// ==============================================================
// Assign command-line options to a cpar_table
// ==============================================================
void _cpar_process_command_line(char *file, uint line, cpar_table *vp, int argc, char *argv[], ...)
{
va_list variable_parameters;
char *key;
gpointer value;
uint type;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

// Phase 1: load up vp with defaults based on the ... arguments to this function
va_start(variable_parameters, argv);
key = va_arg(variable_parameters, char *);
while (key != NULL)
  {
    #ifdef CPAR_DEBUG
    printf("_cpar_process_command_line: Adding key \"%s\"\n", key);
    #endif
    value = g_hash_table_lookup((*vp)->hash_table, key);
    if (value != NULL) _cpar_abort(file, line, "key \"%s\" already exists in parameter table", key);
    type = va_arg(variable_parameters, int);
    switch (type)
      {
        case CPAR_BOOL:
            CPAR_UPDATE(vp, CPAR_ADD_BOOL(key, false));
            break;
        case CPAR_INT:
            CPAR_UPDATE(vp, CPAR_ADD_INT(key, 9999));
            break;
        case CPAR_DOUBLE:
            CPAR_UPDATE(vp, CPAR_ADD_DOUBLE(key, 9E99));
            break;
        case CPAR_REAL:
            CPAR_UPDATE(vp, CPAR_ADD_REAL(key, 9E99));
            break;
        case CPAR_STRING:
            CPAR_UPDATE(vp, CPAR_ADD_STRING(key, "uninitialized"));
            break;
        case CPAR_BOOL_POINTER:
        case CPAR_CHAR_POINTER:
        case CPAR_INT_POINTER:
        case CPAR_DOUBLE_POINTER:
        case CPAR_REAL_POINTER:
        case CPAR_CPAR_POINTER:
        case CPAR_GPOINTER:
            _cpar_abort(file, line, "pointer types aren't allowed for command-line options");
            break;
            
        default:
            _cpar_abort(file, line, "unknown type (%d:%s) for key \"%s\"", type, _cpar_type_name(type), key);
      } // end switch(type)
    if (type != CPAR_BOOL) { CPAR_SET_DEFAULT(vp, key); }
    key = va_arg(variable_parameters, char *);
  } // end while (key != NULL)
va_end(variable_parameters);

// Phase 2: Add things from argv to vp
int i;
char *option_name;
char *option_value = NULL;
uint result = 0;
int temp_int;
double temp_double;
long double temp_real;
// Load program name into special parameter "program_name"
CPAR_UPDATE(vp, CPAR_ADD_STRING("program_name", argv[0]));
for (i = 1; i < argc; i++)
  {
    option_name = argv[i];
    if (option_name[0] != '-')
      {
        // Options are over
        int j;
        uint string_size = 0;
        // get size of remaining options (plus ending \0, plus spaces between options)
        for (j = i; j < argc; j++)
          {
            string_size += strlen(argv[j])+1;
          }
        char *command_string = malloc((uint)string_size+1);
        command_string[0] = '\0';  // Empty string
        // splice options together
        for (j = i; j < argc; j++)
          {
            if (j > i) { strcat(command_string, " "); }  // Space between commands
            strcat(command_string, argv[j]);
          }
        CPAR_UPDATE(vp, CPAR_ADD_STRING("commands", command_string));
        free(command_string);
        // Return index of first command
        CPAR_UPDATE(vp, CPAR_ADD_INT("commands_first_index", i));
        break;
      }
    if (!CPAR_EXISTS(vp, option_name)) 
      {
        _cpar_abort(file, line, "nothing known about this option (key \"%s\")", option_name);
      }
    type = CPAR_GET_TYPE(vp, option_name);
    #ifdef CPAR_DEBUG
    printf("_cpar_process_command_line: Processing command-line argument \"%s\" (type %d:%s)\n", 
           option_name, type, _cpar_type_name(type));
    #endif
    // Get argument after option if needed
    if (type != CPAR_BOOL)
      {
        i++;
        if (i == argc) { _cpar_abort(file, line, "Out of arguments.  No argument for option \"%s\"", 
                                     option_name); }
        option_value = argv[i];
        if (option_value[0] == '-')
          {
            printf("warning: %s should not start with \'-\' (maybe it's not the argument for %s?)\n",
                   option_value, option_name);
          }
        #ifdef CPAR_DEBUG
        printf("_cpar_process_command_line: Processing command-line argument \"%s\"\n", option_value);
        #endif
      }
    switch (type)
      {
        case CPAR_BOOL: // If an option of type BOOL is seen, it is true
            printf("about to overwrite %s\n", option_name);
            CPAR_UPDATE_OVERWRITE(vp, CPAR_ADD_BOOL(option_name, true));
            result = 1;
            break;
        case CPAR_INT:
            result = sscanf(option_value, "%d", &temp_int);
            CPAR_UPDATE(vp, CPAR_ADD_INT(option_name, temp_int));
            break;
        case CPAR_DOUBLE:
            result = sscanf(option_value, "%lf", &temp_double);
            CPAR_UPDATE(vp, CPAR_ADD_DOUBLE(option_name, temp_double));
            break;
        case CPAR_REAL:
            result = sscanf(option_value, "%Lf", &temp_real);
            CPAR_UPDATE(vp, CPAR_ADD_REAL(option_name, (real)temp_real));
            break;
        case CPAR_STRING:
            CPAR_UPDATE(vp, CPAR_ADD_STRING(option_name, option_value));
            result = 1;
            break;

        default:
            _cpar_abort(file, line, "unknown type (%d) for key \"%s\"", type, option_name);
      } // end switch(type)
    if (result != 1) _cpar_abort(file, line, "error converting option \"%s\" value \"%s\" to type (%d:%s)", 
                                 option_name, option_value, type, _cpar_type_name(type)); 
  } // end for i

// Phase 3: Strip out parameters in vp that still have the default flag set
#ifdef CPAR_DEBUG
printf("_cpar_process_command_line: stripping out unused parameters from vp\n");
#endif
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.vp = vp;
g_hash_table_foreach_remove((*vp)->hash_table, _cpar_delete_defaults_GHFunc, (gpointer)(&ghfunc_passer));

// Set all parameters to unseen (except for built-in stuff)
CPAR_CLEAR_ALL_SEEN(vp);
CPAR_SET_SEEN(vp, "program_name");
if (CPAR_EXISTS(vp, "commands")) { CPAR_SET_SEEN(vp, "commands", "commands_first_index"); }
} // end _cpar_process_command_line()


// ==============================================================
// default abort function
// ==============================================================
void _cpar_internal_abort_function(char *file_i, uint line_i, const char *function_i, 
                                   char *file, uint line, char *fmt, ...)
{
va_list args;
va_start(args, fmt);
fprintf(CPAR_ERROR_STREAM, "%s:%d: ", file, line);
vfprintf(CPAR_ERROR_STREAM, fmt, args);
fprintf(CPAR_ERROR_STREAM, "\n(cpar library at:\n%s:%d:%s when error detected)\n", file_i,
        line_i, function_i);
fprintf(CPAR_ERROR_STREAM, "\n");
va_end(args);
fprintf(CPAR_ERROR_STREAM, "Intentionally causing segfault for debugger to catch\n");
fflush(CPAR_ERROR_STREAM);
*((uint *)(0)) = 0;  // trigger a Valgrind stack trace if Valgrind is running

fprintf(CPAR_ERROR_STREAM, "Segfault failed, exiting.\n");
exit(EXIT_FAILURE);
} // end _cpar_internal_abort_function()


// ==============================================================
// Set the abort function (default is _cpar_internal_abort)
// ==============================================================
void cpar_set_abort_function(void (*new_abort_function)(char *file_i, uint line_i, const char *function_i, 
                                                        char *file, uint line, char *fmt, ...))
{
_cpar_abort_function = new_abort_function;
} // end cpar_set_abort_function()


// ==============================================================
// Assign a pointer type to the variable given
// ==============================================================
void _cpar_assign_pointer_type(char *file, uint line, char *type_variable_string, uint *type_variable)
{
_cpar_last_pointer_type++;
*type_variable = _cpar_last_pointer_type;
// Maybe put in fanciers tests based on the string later
} // end _cpar_assign_pointer_type()

// ==============================================================
// Return a string with the type for a CPAR type
// ==============================================================
const char * _cpar_type_name(uint type)
{
if (type >= CPAR_BASIC_TYPES)
  {
    return cpar_type_names[CPAR_UNKNOWN];
  }
else
  {
    return cpar_type_names[type];
  }
} // end _cpar_type_name()

