// Allocation and addition functions
#include <cpar.h>
#include <cpar_internal.h>


// ==============================================================
// Allocate storage for a simple store-by-value parameter
// ==============================================================
gpointer _cpar_allocate_simple_parameter(char *file, uint line, uint size)
{
gpointer memory;
if (size == 0) _cpar_abort(file, line, "size == 0 - internal error");

memory = malloc(size);
if (memory == NULL) _cpar_abort(file, line, "error allocating memory (out of memory?)");
return memory;
} // end _cpar_allocate_simple_parameter()


// ==============================================================
// Copy parameter for a simple store-by-value parameter
// ==============================================================
void _cpar_copy_simple_parameter(char *file, uint line, gpointer to, gpointer from, uint size)
{
uint count;
char *c_from;
char *c_to;
if (size == 0) _cpar_abort(file, line, "size == 0 - internal error");

c_from = (char *)from;
c_to = (char *)to;

for (count = 0; count < size; count++)
  {
    *c_to = *c_from;
    c_from++;
    c_to++;
  }
} // end _cpar_copy_simple_parameter()


// ==============================================================
// Allocate and assign values to a parameter struct
// ==============================================================
cpar_parameter_struct *
_cpar_create_cpar_parameter_struct(char *file, uint line, char *key, 
                                   gpointer pointer_to_value,
                                   cpar_parameter_struct *info)
{
cpar_parameter_struct *new_parameter;
uint copy_type;

// create a new cpar_parameter_struct to hold the parameter
new_parameter = malloc(sizeof(cpar_parameter_struct));
if (new_parameter == NULL)
  {
    _cpar_abort(file, line, "malloc failed (out of memory?)");
  }
new_parameter->type = info->type;
new_parameter->storage = info->storage;
new_parameter->size = info->size;
new_parameter->is_seen = info->is_seen;
new_parameter->is_default = info->is_default;

// If this is a custom gpointer type, treat it like a regular gpointer below
copy_type = info->type;
if (_cpar_type_is_gpointer_custom(info->type)) copy_type = CPAR_GPOINTER;

#ifdef CPAR_DEBUG
printf("_cpar_create_cpar_parameter_struct: adding key: %s, parameter @%lX, type=%d:%s\n",
        key, (unsigned long int)(new_parameter), info->type, _cpar_type_name(info->type));
#endif

// Copy value into parameter_struct (simple store by value)
if (info->storage == CPAR_STORE_BY_VALUE_SIMPLE)
  {
    if (copy_type != CPAR_GPOINTER) _cpar_abort(file, line, 
                                              "store by value used with something other than gpointer type");
    new_parameter->value.gpointer_value = _cpar_allocate_simple_parameter(file, line, info->size);
    _cpar_copy_simple_parameter(file, line, new_parameter->value.gpointer_value,
        (gpointer)(*(double **)pointer_to_value), info->size);
    // See notes in cpar_internal.h for explanation of tricky pointer operations here
  }
  
// Copy value into parameter_struct (complex store by value)
else if (info->storage == CPAR_STORE_BY_VALUE_COMPLEX)
  {
    if (copy_type != CPAR_GPOINTER) _cpar_abort(file, line,
          "store by value used with something other than gpointer type");
    new_parameter->cpar_custom_allocate_function = info->cpar_custom_allocate_function;
    new_parameter->cpar_custom_copy_function = info->cpar_custom_copy_function;
    new_parameter->cpar_custom_free_function = info->cpar_custom_free_function;

    // Copy the structure in pointer_to_value to the new parameter using custom function
    new_parameter->value.gpointer_value = info->cpar_custom_allocate_function();
    info->cpar_custom_copy_function(new_parameter->value.gpointer_value, 
                                    (gpointer)(*(double **)pointer_to_value));
  }

// Copy value into parameter_struct
else
  {
    switch(copy_type)
      {
        case CPAR_BOOL:
          new_parameter->value.bool_value = *((bool *)pointer_to_value);
        break;
        case CPAR_INT:
          new_parameter->value.int_value = *((int *)pointer_to_value);
        break;
        case CPAR_UINT:
          new_parameter->value.uint_value = *((uint *)pointer_to_value);
        break;
        case CPAR_DOUBLE:
          new_parameter->value.double_value = *((double *)pointer_to_value);
        break;
        case CPAR_REAL:
          new_parameter->value.real_value = *((real *)pointer_to_value);
        break;
        case CPAR_BOOL_POINTER:
          new_parameter->value.bool_pointer_value = *((bool **)pointer_to_value);
        break;
        case CPAR_STRING:
          // String is the one case of a "value" type that isn't a value type in C,
          // so extra work has to be done.
          new_parameter->value.char_pointer_value = malloc(strlen(*((char **)pointer_to_value))+1);
          strcpy(new_parameter->value.char_pointer_value, *((char **)pointer_to_value));
        break;
        case CPAR_CHAR_POINTER:
          new_parameter->value.char_pointer_value = *((char **)pointer_to_value);
        break;
        case CPAR_INT_POINTER:
          new_parameter->value.int_pointer_value = *((int **)pointer_to_value);
        break;
        case CPAR_UINT_POINTER:
          new_parameter->value.uint_pointer_value = *((uint **)pointer_to_value);
        break;
        case CPAR_DOUBLE_POINTER:
          new_parameter->value.double_pointer_value = *((double **)pointer_to_value);
        break;
        case CPAR_REAL_POINTER:
          new_parameter->value.real_pointer_value = *((real **)pointer_to_value);
        break;
        case CPAR_CPAR_POINTER:
          new_parameter->value.cpar_pointer_value = *((cpar_table **)pointer_to_value);
        break;
        case CPAR_GPOINTER:
          // See notes in cpar_internal.h for explanation of tricky pointer operations here
          new_parameter->value.gpointer_value = (gpointer)(*(double **)pointer_to_value);
        break;
        default:
              _cpar_abort(file, line, "unknown type: %d", info->type);
        break;
      }
  } // end else not a gpointer store by value type

#ifdef CPAR_DEBUG
// List this parameter if debug is on
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
_cpar_list_parameters_GHFunc(key, (gpointer)(new_parameter), (gpointer)(&ghfunc_passer));
fflush(stdout);
#endif

return new_parameter;
} // end _cpar_create_cpar_parameter_struct()


// ==============================================================
// Create a parameter for transfer from one table to another with a different key
// ==============================================================
void _cpar_transfer_parameter(char *file, uint line, char *to_key,
                              cpar_table *to_vp, char *from_key,
                              cpar_table *from_vp, uint merge_mode)
{
cpar_parameter_struct *from_parameter;

_cpar_assert_cpar_table(file, line, from_vp);
_cpar_assert_cpar_table(file, line, to_vp);

// get parameter to copy
from_parameter = g_hash_table_lookup((*from_vp)->hash_table, from_key);
if (from_parameter == NULL) 
  { 
    // Should we abort if the from_key doesn't exist on a transfer?  I
    // think just ignore is better
    //_cpar_abort(file, line, "key \"%s\" doesn't exist in hash", from_key);
    return;
  }

// Use merge_GHFunc to send this parameter over
ghfunc_passer_struct ghfunc_passer; // defined in this file, only used to pass info to GHFunc
ghfunc_passer.mode = merge_mode;
ghfunc_passer.to_vp = to_vp;
ghfunc_passer.to_key = to_key;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
_cpar_merge_cpar_tables_GHFunc(from_key, from_parameter, (gpointer)(&ghfunc_passer));
} // end _cpar_transfer_parameter()


// ==============================================================
// Add one parameter to a parameter table
// ==============================================================
void _cpar_add_parameter(char *file, uint line, cpar_table *vp, 
                         uint merge_mode, char *key, cpar_parameter_struct *parameter)
{
#ifdef CPAR_DEBUG
printf("_cpar_add_parameter: merge_mode=%d, key=%s\n", merge_mode, key);
#endif
ghfunc_passer_struct ghfunc_passer;
ghfunc_passer.mode = merge_mode;
ghfunc_passer.to_vp = vp;
ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.to_key = NULL; // Just copy keys from old to new
_cpar_assert_cpar_table(file, line, vp);
_cpar_merge_cpar_tables_GHFunc(key, parameter, (gpointer)(&ghfunc_passer));
} // end _cpar_add_parameter()


// ==============================================================
// Add a variable number of parameters to a parameter table
// Format expected in variable parameters:
//
// uint mode
// if mode = CPAR_ADD_PARAMETERS_NORMAL
//    char * key
//    cpar_parameter_struct * new_parameter
// if mode = CPAR_ADD_PARAMETERS_MERGE
//    uint merge_mode
//    cpar_table *merge_me_table
// if mode = CPAR_ADD_PARAMETERS_TRANSFER
//    uint merge_mode
//    char * to_key
//    cpar_table *merge_vp
//    char * from_key
// if mode = CPAR_ADD_PARAMETERS_END
//    end
//
// For merging into an existing table, table_merge_mode is used
// for table merges (CPAR_ADD_MERGE, CPAR_ADD_TRANSFER, and so on), and
// normal_merge_mode is used for CPAR_ADD and CPAR_CREATE_PARAMETERS.
// ==============================================================
void _cpar_add_parameters(char *file, uint line, cpar_table *vp, 
                          uint normal_merge_mode, uint first_add_mode, va_list variable_parameters)
{
uint next_add_mode;
uint table_merge_mode;
char *key;
uint parameter_index = 1;
cpar_table *merge_vp;
cpar_parameter_struct *parameter;

_cpar_assert_cpar_table(file, line, vp);
next_add_mode = first_add_mode;

// While end of parameters list hasn't been reached
while (next_add_mode != CPAR_ADD_PARAMETERS_END)
  {
    #ifdef CPAR_DEBUG
    printf("_cpar_add_parameters: next_add_mode=%d\n", next_add_mode);
    #endif
    switch (next_add_mode)
      {
        case CPAR_ADD_PARAMETERS_NORMAL:
            key = va_arg(variable_parameters, char *);
            parameter = va_arg(variable_parameters, cpar_parameter_struct *);
            _cpar_add_parameter(file, line, vp, normal_merge_mode, key, parameter);
            break;

        case CPAR_ADD_PARAMETERS_MERGE:
            table_merge_mode = va_arg(variable_parameters, int);
            merge_vp = va_arg(variable_parameters, cpar_table *);
            _cpar_merge_cpar_tables(file, line, vp, merge_vp, table_merge_mode);
            break;

        case CPAR_ADD_PARAMETERS_TRANSFER:
            table_merge_mode = va_arg(variable_parameters, int);
            char *to_key = va_arg(variable_parameters, char *);
            merge_vp = va_arg(variable_parameters, cpar_table *);
            char *from_key = va_arg(variable_parameters, char *);
            _cpar_transfer_parameter(file, line, to_key, vp, from_key, merge_vp, table_merge_mode);
            break;

        default:
            _cpar_abort(file, line, "unknown add mode = %d", next_add_mode);
            break;
      } // end switch

    // Each add series should end with a NULL
    key = va_arg(variable_parameters, char *);
    if (key != NULL) { _cpar_abort(file, line, "Each set in a CPAR_ADD list should end with a NULL"); }
    next_add_mode = va_arg(variable_parameters, int);
    parameter_index++;
  } // end while

} // end _cpar_add_parameters()


// ==============================================================
// Add a non-gpointer parameter to a parameter table
// ==============================================================
#define cap(ctype, ctype_name, ctype_preprocessor_name) \
cpar_parameter_struct * _cpar_add_ ## ctype_name(char *file, uint line, char *key, ctype value) \
{ \
cpar_parameter_struct info; \
\
info.type = ctype_preprocessor_name; \
info.storage = CPAR_STORE_UNDEFINED; \
info.is_seen = false; \
info.is_default = false; \
return _cpar_create_cpar_parameter_struct(file, line, key, (gpointer)(&value), &info); \
}

cap(bool, bool, CPAR_BOOL)
cap(int, int, CPAR_INT)
cap(uint, uint, CPAR_UINT)
cap(double, double, CPAR_DOUBLE)
cap(real, real, CPAR_REAL)
cap(bool *, bool_pointer, CPAR_BOOL_POINTER)
cap(char *, string, CPAR_STRING)
cap(char *, char_pointer, CPAR_CHAR_POINTER)
cap(int *, int_pointer, CPAR_INT_POINTER)
cap(uint *, uint_pointer, CPAR_UINT_POINTER)
cap(double *, double_pointer, CPAR_DOUBLE_POINTER)
cap(real *, real_pointer, CPAR_REAL_POINTER)
cap(cpar_table *, cpar_pointer, CPAR_CPAR_POINTER)


// ==============================================================
// Add a non-gpointer parameter to a parameter table (type given in parameters)
// ==============================================================
cpar_parameter_struct * _cpar_add(char *file, uint line, char *key, uint type, gpointer value) 
{ 
cpar_parameter_struct info; 

info.type = type;
info.storage = CPAR_STORE_UNDEFINED; 
info.size = 0;  // Not used
info.is_seen = false;
info.is_default = false;
return _cpar_create_cpar_parameter_struct(file, line, key, value, &info); 
}


// ==============================================================
// Add a gpointer parameter to a parameter table (store by reference)
// ==============================================================
cpar_parameter_struct * _cpar_add_gpointer(char *file, uint line, char *key, 
                                           uint gpointer_type, gpointer value)
{
cpar_parameter_struct info;

info.type = gpointer_type;
info.storage = CPAR_STORE_BY_REFERENCE;
info.size = 0;  // Not used
info.is_seen = false;
info.is_default = false;
return _cpar_create_cpar_parameter_struct(file, line, key, &value, &info); 
}

// ==============================================================
// Add a gpointer parameter to a parameter table (store by value - simple)
// ==============================================================
cpar_parameter_struct * _cpar_add_sv_gpointer(char *file, uint line, char *key, 
                                              uint gpointer_type, gpointer value, uint size)
{
cpar_parameter_struct info;

info.type = gpointer_type;
info.storage = CPAR_STORE_BY_VALUE_SIMPLE;
info.size = size;
info.is_seen = false;
info.is_default = false;
return _cpar_create_cpar_parameter_struct(file, line, key, &value, &info);
}

// ==============================================================
// Add a gpointer parameter to a parameter table (store by value - complex)
// ==============================================================
cpar_parameter_struct * _cpar_add_cv_gpointer(char *file, uint line, char *key, 
                                              uint gpointer_type, gpointer value,
        CPAR_COMPLEX_ALLOCATE_PROTOTYPE, CPAR_COMPLEX_COPY_PROTOTYPE, CPAR_COMPLEX_FREE_PROTOTYPE)
{
cpar_parameter_struct info;

info.type = gpointer_type;
info.storage = CPAR_STORE_BY_VALUE_COMPLEX;
info.size = 0;  // Not used
info.is_seen = false;
info.is_default = false;
info.cpar_custom_allocate_function = cpar_custom_allocate_function;
info.cpar_custom_copy_function = cpar_custom_copy_function;
info.cpar_custom_free_function = cpar_custom_free_function;
return _cpar_create_cpar_parameter_struct(file, line, key, &value, &info);
}


void _cpar_destroy_key(gpointer key);
void _cpar_destroy_value(gpointer value);
// ==============================================================
// Return an empty cpar table (type cpar_table*)
// ==============================================================
cpar_table * _cpar_create_empty_cpar_table(char *file, uint line)
{
cpar_table *vp;

// Create pointer to hash table
vp = malloc(sizeof(struct cpar_table_struct *));
if (vp == NULL)
  {
    _cpar_abort(file, line, "malloc() failed (out of memory?)");
  }

// Create hash table
*vp = malloc(sizeof(struct cpar_table_struct));
if (*vp == NULL)
  {
    _cpar_abort(file, line, "malloc() failed (out of memory?)");
  }
(*vp)->magic_key = _CPAR_MAGIC_KEY;
(*vp)->hash_table = g_hash_table_new_full(g_str_hash, g_str_equal, _cpar_destroy_key, _cpar_destroy_value);
if ((*vp)->hash_table == NULL)
  {
    _cpar_abort(file, line, "g_hash_table_new() failed (out of memory?)");
  }

return vp;
} // end _cpar_create_empty_cpar_table()


// ==============================================================
// Create a cpar table (type cpar_type*) and insert keys and values
// corresponding to function parameters.  The parameters to this
// function should be a list of the pair:
// (key, cpar parameter struct), ending with a NULL
// ==============================================================
cpar_table * _cpar_create_cpar_table(char *file, uint line, uint first_add_mode, ...)
{
va_list variable_parameters;
cpar_table *vp;

// Create cpar table
vp = _cpar_create_empty_cpar_table(file, line);

// Pass variable parameters on to _cpar_add_parameters()
va_start(variable_parameters, first_add_mode);
_cpar_add_parameters(file, line, vp, CPAR_MERGE_OVERWRITE_MODE, first_add_mode, variable_parameters);
va_end(variable_parameters);
return vp;
} // end _cpar_create_cpar_table()


// ==============================================================
// Add keys and values to an already existing hash table
// Almost the same as cparameters_create_variable_parameters_hash()
// above, it just doesn't create the hash.
// ==============================================================
void _cpar_add_parameters_macro(char *file, uint line, cpar_table *vp, 
                                uint merge_mode, uint first_add_mode, ...)
{
va_list variable_parameters;

if (vp == NULL) return;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

// Pass variable parameters on to _cpar_add_parameters()
va_start(variable_parameters, first_add_mode);
_cpar_add_parameters(file, line, vp, merge_mode, first_add_mode, variable_parameters);
va_end(variable_parameters);
} // end _cpar_add_parameters_macro()


// ==============================================================
// Free the value of a key-value pair
// ==============================================================
bool _cpar_free_parameter(gpointer key, gpointer value, gpointer user_data)
{
// Don't free the key, this is the responsibility of the user
// Free the value gpointer though g_hash_table_foreach() in the function
// cpar_destroy_cpar_table_hash() below
ghfunc_passer_struct *ghfunc_passer = (ghfunc_passer_struct *)user_data;
cpar_table *vp = ghfunc_passer->vp;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(ghfunc_passer->file, ghfunc_passer->line, vp);

#ifdef CPAR_DEBUG
printf("about to free:\n");
_cpar_list_parameters_GHFunc(key, value, user_data);
#endif

_cpar_destroy_value(value);
return true;
} // end _cpar_free_parameter()


// ==============================================================
// Free a key (used internally by glib)
// ==============================================================
void _cpar_destroy_key(gpointer key)
{
if (key == NULL) _cpar_abort("unknown", 0, "_cpar_destroy_key(): key is NULL");

free(key);
} // end _cpar_destroy_key()


// ==============================================================
// Free a hash table value (which is a parameter for us) (used internally by glib)
// ==============================================================
void _cpar_destroy_value(gpointer value)
{
cpar_parameter_struct *parameter = (cpar_parameter_struct *)value;
if (value == NULL) _cpar_abort("unknown", 0, "_cpar_destroy_value(): value is NULL");

// We allocated a string for the CPAR_STRING type
if (parameter->type == CPAR_STRING)
  {
    #ifdef CPAR_DEBUG
    printf("freeing string: %s\n", parameter->value.char_pointer_value);
    #endif
    free(parameter->value.char_pointer_value);
  }

// Free the value storage if this is a simple store by value gpointer
if (_cpar_type_is_gpointer_custom(parameter->type) && (parameter->storage == CPAR_STORE_BY_VALUE_SIMPLE))
  {
    #ifdef CPAR_DEBUG
    printf("freeing simple store by value @ %lX\n", (unsigned long int)(parameter->value.gpointer_value));
    #endif
    free(parameter->value.gpointer_value);
  }

// Free the value storage if this is a complex store by value gpointer
if (_cpar_type_is_gpointer_custom(parameter->type) && (parameter->storage == CPAR_STORE_BY_VALUE_COMPLEX))
  {
    #ifdef CPAR_DEBUG
    printf("freeing complex store by value @ %lX\n", (unsigned long int)(parameter->value.gpointer_value));
    #endif
    parameter->cpar_custom_free_function(parameter->value.gpointer_value);
  }

free(parameter);
} // end _cpar_destroy_value()


// ==============================================================
// Free a parameters hash table
// ==============================================================
void _cpar_destroy_cpar_table(char *file, uint line, cpar_table *vp)
{
ghfunc_passer_struct ghfunc_passer;
if (vp == NULL) return;

// Make sure vp is really a cpar_table
_cpar_assert_cpar_table(file, line, vp);

// Overwrite magic key to detect double frees
(*vp)->magic_key = _CPAR_MAGIC_KEY_DELETED;

ghfunc_passer.file = file;
ghfunc_passer.line = line;
ghfunc_passer.vp = vp;
g_hash_table_destroy((*vp)->hash_table);
free(*vp);
} // end _cpar_destroy_cpar_table()

